#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include <vector>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
#include <io.h>
#include <direct.h>
#include <errno.h>
#else // !_WIN32
#include <unistd.h>
#endif // _WIN32
#include "Source/JPEG.h"
#include "Source/PNG.h"
#include "Source/TIFF.h"
#include "Source/CmdLineParser.h"
#include "Source/Array.h"
#include "Source/Util.h"

#define PARALLEL_IO_CODE 0		// For the xy smoothing, read in the blocks in parallel
#define LAZY_IO 0				// For the z smoothing, using pre-fetching and lazy write-backs for the I/O

const char DEFAULT_EXTENSION[] = "tiff";
typedef float FLT;

#ifdef _WIN32
const char FILE_SEPARATOR = '\\';
#else // !_WIN32
const char FILE_SEPARATOR = '/';
#endif // _WIN32
bool ExistsDir( const char* dirName )
{
    struct stat buf;
	char _dirName[512];
	strcpy( _dirName , dirName );
	if( _dirName[ strlen(_dirName)-1 ]==FILE_SEPARATOR ) _dirName[ strlen(_dirName)-1 ] = 0 ;

	if( stat( _dirName , &buf )!=0 ) return false;
    else if( buf.st_mode & S_IFDIR ) return true;
    else return false;
}
#ifdef _WIN32
bool RemoveDir( const char* dirName ){ return _rmdir( dirName )==0; }
bool   MakeDir( const char* dirName ){ return ExistsDir( dirName ) || ( _mkdir( dirName )==0 ); }
#else // !_WIN32
bool RemoveDir( const char* dirName ){ return rmdir( dirName )==0; }
bool   MakeDir( const char* dirName ){ return ExistsDir( dirName ) || ( mkdir( dirName , S_IRWXU | S_IXGRP | S_IXOTH )==0 ); }
#endif // _WIN32

cmdLineString In( "in" ) , Out( "out" ) , OutHeader( "header" ) , OutExtension( "ext" ) , ScratchHeader( "scratchHeader" );
cmdLineFloat XYSigma( "xySigma" , 0.f ) , ZSigma( "zSigma" , 0.f ) , ZRejectDeviation( "zReject" , 2.f );
cmdLineInt Channels( "channels" , 1 ) , Threads( "threads" , 1 ) , XYBlocks( "xyBlocks" , 4 );
cmdLineReadable Verbose( "verbose" ) , Progress( "progress" ) , NoCleanUp( "noCleanUp" ) , NoBlack( "noBlack" ) , MakeSubDirs( "subDirs" );
cmdLineReadable* params[] = { &In , &Out , &XYSigma , &ZSigma , &OutHeader , &ScratchHeader , &OutExtension , &Channels , &Threads , &Verbose , &NoCleanUp , &Progress , &ZRejectDeviation , &NoBlack , &MakeSubDirs , &XYBlocks };

void ShowUsage( const char* ex )
{
	printf( "Usage %s:\n" , ex );
	printf( "\t --%s <input image file>\n" , In.name );
	printf( "\t[--%s <output file list>]\n" , Out.name );
	printf( "\t[--%s <output file header>]\n" , OutHeader.name );
	printf( "\t[--%s <output file extension>=%s]\n" , OutExtension.name , DEFAULT_EXTENSION );
	printf( "\t[--%s <temporary file header>]\n" , ScratchHeader.name );
	printf( "\t[--%s <in-slice gaussian standard deviation>=%f]\n" , XYSigma.name , XYSigma.value );
	printf( "\t[--%s <cross-slice gaussian standard deviation>=%f]\n" , ZSigma.name , ZSigma.value );
	printf( "\t[--%s <parallelization threads>=%d]\n" , Threads.name , Threads.value );
	printf( "\t[--%s <output channels>=%d]\n" , Channels.name , Channels.value );
	printf( "\t[--%s <z-rejection standard deviation>]\n" , ZRejectDeviation.name );
	printf( "\t[--%s <number of blocks for xy-smoothing>=%d]\n" , XYBlocks.name , XYBlocks.value );
	printf( "\t[--%s]\n" , MakeSubDirs.name );
	printf( "\t[--%s]\n" , NoBlack.name );
	printf( "\t[--%s]\n" , NoCleanUp.name );
	printf( "\t[--%s]\n" , Progress.name );
	printf( "\t[--%s]\n" , Verbose.name );
}

template< int SIZE=512 >
class MyString
{
	char _string[SIZE];
public:
	MyString( void ){ _string[0]=0; }
	MyString( const MyString& str ){ strcpy( _string , str._string ); }
	MyString( const char* s )
	{
		if( strlen(s)>=SIZE-1 ) fprintf( stderr , "[ERROR] MyString::MyString: file-name is too long: %d>=%d\n" , strlen(s) , SIZE-1 ) , exit( 0 );
		strcpy( _string , s );
	}
	MyString& operator = ( const MyString& str )
	{
		strcpy( _string , str._string );
		return *this;
	}
	MyString& operator = ( const char* s )
	{
		if( strlen(s)>=SIZE-1 ) fprintf( stderr , "[ERROR] MyString::operator =: file-name is too long: %d>=%d\n" , strlen(s) , SIZE-1 ) , exit( 0 );
		strcpy( _string , s );
		return *this;
	}
	char* operator()( void ) { return _string; }
	const char* operator()( void ) const { return _string; }
};

template< class Real , int Channels >
class Pixel
{
	Real _p[Channels];
public:
	Pixel( void ){ for( int c=0 ; c<Channels ; c++ ) _p[c] = (Real)0; }
	Real& operator[] ( int c ){ return _p[c]; }
	const Real& operator[] ( int c ) const { return _p[c]; }
	bool operator == ( const Pixel& p ) const
	{
		for( int c=0 ; c<Channels ; c++ ) if( _p[c]!=p._p[c] ) return false;
		return true;
	}
	bool operator != ( const Pixel& p ) const
	{
		for( int c=0 ; c<Channels ; c++ ) if( _p[c]!=p._p[c] ) return true;
		return false;
	}
	Pixel operator - ( const Pixel& p ) const { Pixel pp ; for( int c=0 ; c<Channels ; c++ ) pp._p[c] = _p[c] - p._p[c] ; return pp; }
	Pixel operator + ( const Pixel& p ) const { Pixel pp ; for( int c=0 ; c<Channels ; c++ ) pp._p[c] = _p[c] + p._p[c] ; return pp; }
	Pixel operator * ( Real s ) const { Pixel pp ; for( int c=0 ; c<Channels ; c++ ) pp._p[c] = _p[c] * s ; return pp; }
	Pixel operator / ( Real s ) const { Pixel pp ; for( int c=0 ; c<Channels ; c++ ) pp._p[c] = _p[c] / s ; return pp; }

	Pixel& operator -= ( const Pixel& p ){ for( int c=0 ; c<Channels ; c++ ) _p[c] -= p._p[c] ; return *this; }
	Pixel& operator += ( const Pixel& p ){ for( int c=0 ; c<Channels ; c++ ) _p[c] += p._p[c] ; return *this; }
	Pixel& operator *= ( Real s ){ for( int c=0 ; c<Channels ; c++ ) _p[c] *= s ; return *this; }
	Pixel& operator /= ( Real s ){ for( int c=0 ; c<Channels ; c++ ) _p[c] /= s ; return *this; }
};
template< int Channels >
Pixel< FLT , Channels > ToFloat( Pixel< unsigned char , Channels > p )
{
	Pixel< FLT , Channels > _p;
	for( int c=0 ; c<Channels ; c++ ) _p[c] = (FLT)p[c];
	return _p;
}
template< int Channels >
Pixel< unsigned char , Channels > FromFloat( Pixel< FLT , Channels > p ) 
{
	Pixel< unsigned char , Channels > _p;
	for( int c=0 ; c<Channels ; c++ ) _p[c] = (unsigned char)( std::max< int >( 0 , std::min< int >( 255 , (int)floor( p[c] + (FLT)0.5 ) ) ) );
	return _p;
}
template< int Channels > FLT SquareNorm( Pixel< FLT , Channels > p );
template<> FLT SquareNorm( Pixel< FLT , 1 > p ){ return p[0] * p[0]; }
template<> FLT SquareNorm( Pixel< FLT , 3 > p ){ return p[0] * p[0]  + p[1] * p[1] + p[2] * p[2]; }
template< int Channels >
FLT SquareNorm( Pixel< FLT , Channels > p )
{
	FLT n2 = 0;
	for( int c=0 ; c<Channels ; c++ ) n2 += p[c] * p[c];
	return n2;
}

void ImageInfo( const char* fileName , int& width , int& height , int& channelsPerPixel , int& bitsPerChannel )
{
	char* fileExtension = GetFileExtension( fileName );
	if     ( !strcasecmp( fileExtension , "jpg" ) || !strcasecmp( fileExtension , "jpeg" ) ) JPEGGetImageSize( fileName , width , height , channelsPerPixel , bitsPerChannel );
	else if( !strcasecmp( fileExtension , "tif" ) || !strcasecmp( fileExtension , "tiff" ) ) TIFFGetImageSize( fileName , width , height , channelsPerPixel , bitsPerChannel );
	else if( !strcasecmp( fileExtension , "png" ) )                                           PNGGetImageSize( fileName , width , height , channelsPerPixel , bitsPerChannel );
	else fprintf( stderr , "[ERROR] Unrecognized file extension: %s (%s)\n" , fileExtension , fileName ) , exit( 0 );
	delete[] fileExtension;
}

template< int Channels >
Pointer( Pixel< unsigned char , Channels > ) ReadImage( const char* fileName , int& width , int& height )
{
	if( Channels!=1 && Channels!=3 ) fprintf( stderr , "[ERROR] ReadImage: Only 1 and 3 channel input supported\n" ) , exit( 0 );
	Pointer( unsigned char ) pixels = NullPointer< unsigned char >();
	char* fileExtension = GetFileExtension( fileName );
	if     ( !strcasecmp( fileExtension , "jpg" ) || !strcasecmp( fileExtension , "jpeg" ) ) pixels = Channels==1 ? JPEGReadGray( fileName , width , height ) : JPEGReadColor( fileName , width , height );
	else if( !strcasecmp( fileExtension , "tif" ) || !strcasecmp( fileExtension , "tiff" ) ) pixels = Channels==1 ? TIFFReadGray( fileName , width , height ) : TIFFReadColor( fileName , width , height );
	else if( !strcasecmp( fileExtension , "png" ) )                                          pixels = Channels==1 ?  PNGReadGray( fileName , width , height ) :  PNGReadColor( fileName , width , height );
	else fprintf( stderr , "[ERROR] Unrecognized file extension: %s (%s)\n" , fileExtension , fileName ) , exit( 0 );
	delete[] fileExtension;
	if( !pixels ) fprintf( stderr , "[ERROR] Failed to read pixels: %s\n" , fileName ) , exit( 0 );
	return ( Pointer( Pixel< unsigned char , Channels > ) )pixels;
}

template< int Channels >
void ReadImage( const char* fileName , Pointer( unsigned char ) pixels , int width , int height )
{
	if( Channels!=1 && Channels!=3 ) fprintf( stderr , "[ERROR] ReadImage: Only 1 and 3 channel input supported\n" ) , exit( 0 );
	char* fileExtension = GetFileExtension( fileName );
	if     ( !strcasecmp( fileExtension , "jpg" ) || !strcasecmp( fileExtension , "jpeg" ) ) Channels==1 ? JPEGReadGray( fileName , pixels , width , height ) : JPEGReadColor( fileName , pixels , width , height );
	else if( !strcasecmp( fileExtension , "tif" ) || !strcasecmp( fileExtension , "tiff" ) ) Channels==1 ? TIFFReadGray( fileName , pixels , width , height ) : TIFFReadColor( fileName , pixels , width , height );
	else if( !strcasecmp( fileExtension , "png" ) )                                          Channels==1 ?  PNGReadGray( fileName , pixels , width , height ) :  PNGReadColor( fileName , pixels , width , height );
	else fprintf( stderr , "[ERROR] Unrecognized file extension: %s (%s)\n" , fileExtension , fileName ) , exit( 0 );
	delete[] fileExtension;
}

template< int Channels >
void WriteImage( const char* fileName , ConstPointer( Pixel< unsigned char , Channels > ) pixels , int width , int height )
{
	if( Channels!=1 && Channels!=3 ) fprintf( stderr , "[ERROR] WriteImage: Only 1 and 3 channel input supported\n" ) , exit( 0 );
	char* fileExtension = GetFileExtension( fileName );
	ConstPointer( unsigned char ) _pixels = ( ConstPointer( unsigned char ) )pixels;
	if     ( !strcasecmp( fileExtension , "jpg" ) || !strcasecmp( fileExtension , "jpeg" ) ) Channels==1 ? JPEGWriteGray( fileName , _pixels , width , height , 100 ) : JPEGWriteColor( fileName , _pixels , width , height , 100 );
	else if( !strcasecmp( fileExtension , "tif" ) || !strcasecmp( fileExtension , "tiff" ) ) Channels==1 ? TIFFWriteGray( fileName , _pixels , width , height       ) : TIFFWriteColor( fileName , _pixels , width , height       );
	else if( !strcasecmp( fileExtension , "png" ) )                                          Channels==1 ?  PNGWriteGray( fileName , _pixels , width , height       ) :  PNGWriteColor( fileName , _pixels , width , height       );
	else fprintf( stderr , "[ERROR] Unrecognized file extension: %s (%s)\n" , fileExtension , fileName ) , exit( 0 );
	delete[] fileExtension;
}


struct ImageTileGrid2D
{
protected:
	Pointer( int ) _rowHeights;
	Pointer( int ) _columnWidths;
	int _width , _height;
	bool _gray;
public:
	char** fileNames;
	int columns , rows;
	bool   gray( void )  const { return _gray; }
	int   width( void )  const { return _width; }
	int  height( void )  const { return _height; }
	int  bWidth( void )  const { return _columnWidths[0]; }
	int bHeight( void )  const { return _rowHeights[0]; }
	int  cWidth( int c ) const { return _columnWidths[c]; }
	int rHeight( int r ) const { return _rowHeights[r]; }
	ImageTileGrid2D( void ){ columns = rows = _width = _height = 0 , _gray = false , _rowHeights = _columnWidths = NullPointer< int >() , fileNames = NULL; }
	ImageTileGrid2D( int c , int r , char** const fNames ) : columns(c) , rows(r)
	{
		fileNames = new char*[columns*rows];
		if( !fileNames ) fprintf( stderr , "[ERROR] ImageTileGrid2D: Failed to allocate tiles\n" ) , exit( 0 );
		for( int i=0 ; i<columns*rows ; i++ )
		{
			fileNames[i] = new char[ strlen( fNames[i] ) + 1 ];
			strcpy( fileNames[i] , fNames[i] );
		}
		_rowHeights = NewPointer< int >( rows ) , _columnWidths = NewPointer< int >( columns );

		_gray = true;
		_width = _height = 0;
		{
			int w , h , cpp , bpc;
			for( int j=0 ; j<rows ; j++ ) for( int i=0 ; i<columns ; i++ )
			{
				ImageInfo( fNames[j*columns+i] , w , h , cpp , bpc );
				if( !i ) _rowHeights[j] = h , _height += h;
				else if( _rowHeights[j]!=h ) fprintf( stderr , "[ERROR] ImageTileGrid2D: row height is inconsistent: %d != %d (%d,%d)\n" , _rowHeights[j] , h , i , j ) , exit( 0 );
				if( !j ) _columnWidths[i] = w , _width += w;
				else if( _columnWidths[i]!=w ) fprintf( stderr , "[ERROR] ImageTileGrid2D: column widths are inconsistent: %d != %d (%d,%d)\n" , _columnWidths[i] , w , i , j ) , exit( 0 );
				if( cpp!=1 ) _gray = false;
			}
		}
		// [STRICT BLOCKS]
		// Test that rows have the same height (except perhaps the last one)
		for( int j=0 ; j<rows-1; j++ ) if( _rowHeights[j]!=_rowHeights[0] ) fprintf( stderr , "[ERROR] ImageTileGrid2D: row heights differ: %d != %d (%d)\n" , _rowHeights[0] , _rowHeights[j] , j ) , exit( 0 );
		if( _rowHeights[rows-1]>_rowHeights[0] ) fprintf( stderr , "[ERROR] ImageTileGrid2D: row heights differ: %d > %d\n" , _rowHeights[0] , _rowHeights[rows-1] ) , exit( 0 );
		// Test that columns have the same width (except perhaps the last one)
		for( int i=0 ; i<columns-i; i++ ) if( _columnWidths[i]!=_columnWidths[0] ) fprintf( stderr , "[ERROR] ImageTileGrid2D: column widths differ: %d != %d (%d)\n" , _columnWidths[0] , _columnWidths[i] , i ) , exit( 0 );
		if( _columnWidths[columns-1]>_columnWidths[0] ) fprintf( stderr , "[ERROR] ImageTileGrid2D: column widths differ: %d > %d\n" , _columnWidths[0] , _columnWidths[columns-1] ) , exit( 0 );
	}
	~ImageTileGrid2D( void )
	{
		for( int i=0 ; i<columns*rows ; i++ ) delete[] fileNames[i];
		delete[] fileNames;
		fileNames = NULL;
		if( _rowHeights ) DeletePointer( _rowHeights );
		if( _columnWidths ) DeletePointer( _columnWidths );
		columns = rows = 0 , _rowHeights = _columnWidths = NullPointer< int >();
	}
	void removeFiles( void )
	{
		if( fileNames )
		{
			for( int i=0 ; i<columns*rows ; i++ )
			{
				remove( fileNames[i] );
				delete[] fileNames[i];
			}
			delete[] fileNames;
			fileNames = NULL;
		}
		if( _rowHeights ) DeletePointer( _rowHeights );
		if( _columnWidths ) DeletePointer( _columnWidths );
		columns = rows = 0 , _rowHeights = _columnWidths = NullPointer< int >();
	}
	void readBlock( Pointer( unsigned char ) pixels , int c , int r , bool gray ) const
	{
		if( c<0 || c>=columns || r<0 || r>=rows ) return;

		int channels = gray ? 1 : 3;
		int w , h;
		Pointer( unsigned char ) _pixels = gray ? ( Pointer( unsigned char ) )ReadImage< 1 >( fileNames[r*columns+c] , w , h ) : ( Pointer( unsigned char ) )ReadImage< 3 >( fileNames[r*columns+c] , w , h );
		if( w!=bWidth() ) for( int j=0 ; j<h ; j++ ) memcpy( pixels + j * bWidth() * channels , _pixels + j * w * channels , sizeof(unsigned char) * w * channels );
		else memcpy( pixels , _pixels , sizeof(unsigned char) * w * h * channels );
		DeletePointer( _pixels );
	}
};


struct ImageTileGrid3D
{
protected:
	int _layers;
	int _width , _height;
	bool _gray;
	Pointer( ImageTileGrid2D* ) _imageSlices;
public:
	int layers ( void )  const { return _layers; }
	bool gray  ( void )  const { return _gray; }
	int width  ( void )  const { return _width; }
	int height ( void )  const { return _height; }
	int columns( void )  const { return _imageSlices[0]->columns; }
	int rows   ( void )  const { return _imageSlices[0]->rows; }
	int  bWidth( void )  const { return _imageSlices[0]->bWidth(); }
	int bHeight( void )  const { return _imageSlices[0]->bHeight(); }
	int  cWidth( int c ) const { return _imageSlices[0]->cWidth(c); }
	int rHeight( int r ) const { return _imageSlices[0]->rHeight(r); }
	ImageTileGrid3D( void ){ _layers = 0 , _imageSlices = NullPointer< ImageTileGrid2D* >(); }
	ImageTileGrid3D( const char* fileName )
	{
		int columns , rows;
		columns = rows = _layers = 0;
		_imageSlices = NullPointer< ImageTileGrid2D* >();

		FILE* fp = fopen( fileName , "r" );
		if( !fp ) fprintf( stderr , "[ERROR] ImageTileGrid3D: Failed to open file for reading: %s\n" , fileName ) , exit( 0 );
		char headers[3][512];
		if( !fgets( headers[0] , 511 , fp ) || !fgets( headers[1] , 511 , fp ) || !fgets( headers[2] , 511 , fp ) )
			fprintf( stderr , "[ERROR] Failed to read file header\n" ) , exit( 0 );
		for( int i=0 ; i<3 ; i++ ) sscanf( headers[i] , " width: %d " , &columns ) , sscanf( headers[i] , " height: %d " , &rows ) , sscanf( headers[i] , " depth: %d " , &_layers );
		if( columns<=0 || rows<=0 || _layers<=0 ) fprintf( stderr , "[ERROR] ImageTileGrid3D: Failed to read tile dimensions: %d x %d x %d\n" , columns , rows , _layers ) , exit( 0 );
		_imageSlices = NewPointer< ImageTileGrid2D* >( _layers );
		char** fileNames = new char*[ columns * rows ];
		for( int d=0 ; d<_layers ; d++ )
		{
			if( Progress.set ) printf( "Initializing Layer: %d / %d               \r" , d , _layers );
			char _fileName[1024] , __fileName[1024];
			for( int i=0 ; i<columns*rows ; i++ )
			{
				if( !fgets( _fileName , 1023 , fp ) ) fprintf( stderr , "[ERROR] ImageTileGrid3D: Failed to read file-name[%d %d]\n" , i , d ) , exit( 0 );
				sscanf( _fileName , " %s " , __fileName );
				fileNames[i] = new char[ strlen( __fileName ) + 1 ];
				strcpy( fileNames[i] , __fileName );
			}
			_imageSlices[d] = new ImageTileGrid2D( columns , rows , fileNames );
			for( int i=0 ; i<columns*rows ; i++ ) delete[] fileNames[i];
		}
		if( Progress.set ) printf( "\n" );
		fclose( fp );
		delete[] fileNames;
		_width = _imageSlices[0]->width() , _height = _imageSlices[0]->height();
		_gray = true;
		for( int d=0 ; d<_layers ; d++ )
		{
			if( !_imageSlices[d]->gray() ) _gray = false;
			if( _width!=_imageSlices[d]->width() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: total width is inconsistent: %d != %d (%d)\n" , _width , _imageSlices[d]->width() , d ) , exit( 0 );
			if( _height!=_imageSlices[d]->height() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: total height is inconsistent: %d != %d (%d)\n" , _height , _imageSlices[d]->height() , d ) , exit( 0 );
			if( _imageSlices[d]->bWidth()!=_imageSlices[0]->bWidth() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: column widths are inconsistent: %d != %d (%d)\n" , _imageSlices[0]->bWidth() , _imageSlices[d]->bWidth() , d ) , exit( 0 );
			if( _imageSlices[d]->bHeight()!=_imageSlices[0]->bHeight() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: row heights are inconsistent: %d != %d (%d)\n" , _imageSlices[0]->bHeight() , _imageSlices[d]->bHeight() , d ) , exit( 0 );
		}
	}
	ImageTileGrid3D( char** const fileNames , int columns , int rows , int layers )
	{
		_layers = layers;
		_imageSlices = NewPointer< ImageTileGrid2D* >( _layers );
		for( int d=0 ; d<_layers ; d++ ) _imageSlices[d] = new ImageTileGrid2D( columns , rows , fileNames+columns*rows*d );
		_width = _imageSlices[0]->width() , _height = _imageSlices[0]->height();
		_gray = true;
		for( int d=0 ; d<_layers ; d++ )
		{
			if( !_imageSlices[d]->gray() ) _gray = false;
			if( _width!=_imageSlices[d]->width() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: total width is inconsistent: %d != %d (%d)\n" , _width , _imageSlices[d]->width() , d ) , exit( 0 );
			if( _height!=_imageSlices[d]->height() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: total height is inconsistent: %d != %d (%d)\n" , _height , _imageSlices[d]->height() , d ) , exit( 0 );
			if( _imageSlices[d]->bWidth()!=_imageSlices[0]->bWidth() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: column widths are inconsistent: %d != %d (%d)\n" , _imageSlices[0]->bWidth() , _imageSlices[d]->bWidth() , d ) , exit( 0 );
			if( _imageSlices[d]->bHeight()!=_imageSlices[0]->bHeight() ) fprintf( stderr , "[ERROR] ImageTileGrid3D: row heights are inconsistent: %d != %d (%d)\n" , _imageSlices[0]->bHeight() , _imageSlices[d]->bHeight() , d ) , exit( 0 );
		}
	}
	~ImageTileGrid3D( void )
	{
		if( _imageSlices )
		{
			for( int d=0 ; d<_layers ; d++ ) if( _imageSlices[d] ) delete _imageSlices[d];
			DeletePointer( _imageSlices );
		}
		_layers = 0;
		_imageSlices = NullPointer< ImageTileGrid2D* >();
	}
	void removeFiles( void )
	{
		if( _imageSlices )
		{
			for( int d=0 ; d<_layers ; d++ ) if( _imageSlices[d] ) { _imageSlices[d]->removeFiles() ; delete _imageSlices[d]; }
			DeletePointer( _imageSlices );
		}
		_layers = 0;
		_imageSlices = NullPointer< ImageTileGrid2D* >();
	}

	void readBlock( Pointer( unsigned char ) pixels , int c , int r , int d , bool gray ) const
	{
		if( d<0 || d>=_layers ) return;
		_imageSlices[d]->readBlock( pixels , c , r , gray );
	}
	ImageTileGrid2D& operator[] ( int layer ){ return *_imageSlices[layer]; }
	const ImageTileGrid2D& operator[] ( int layer ) const { return *_imageSlices[layer]; }
};
template< int Channels >
ImageTileGrid3D* smoothXY( const ImageTileGrid3D* in )
{
	typedef Pixel< unsigned char , Channels > PixelI;
	typedef Pixel<           FLT , Channels > PixelF;

	float xyDeviation = XYSigma.value;
	int xyRadius = (int)ceil( XYSigma.value*3 );
	FLT* weights = new FLT[xyRadius*2+1];

	if( xyRadius+1>in->bWidth() || xyRadius+1>in->bHeight() ) fprintf( stderr , "[ERROR] Block size is too small: (%d,%d) > (%d,%d)\n" , xyRadius+1 , xyRadius+1 , in->bWidth() , in->bHeight() ) , exit( 0 );
	int bWidth = in->bWidth() , bHeight = in->bHeight();

	{
		FLT sum = 0;
		for( int i=-xyRadius ; i<=xyRadius ; i++ ) weights[i+xyRadius] = (FLT)exp( -(double)(i*i)/( 2. * xyDeviation * xyDeviation ) ) , sum += weights[i+xyRadius];
		for( int i=-xyRadius ; i<=xyRadius ; i++ ) weights[i+xyRadius] /= sum;
	}

	int xBlocks = in->columns() , yBlocks = in->rows();
	int yMultiBlocks = ( yBlocks + XYBlocks.value - 1 ) / XYBlocks.value;
	if( Verbose.set ) printf( "\t\tBlock Size: %d x %d\n" , bWidth , bHeight ) , printf( "\t\tBlocks: %d x %d\n" , xBlocks , yBlocks );

	Pointer( Pointer( PixelI ) ) _inBlocks[3];
	Pointer( Pointer( PixelI ) )  _xBlocks;
	Pointer( Pointer( PixelI ) ) _xyBlocks;
	for( int i=0 ; i<3 ; i++ )
	{
		_inBlocks[i] = AllocPointer< Pointer( PixelI ) >( XYBlocks.value+2 );
		if( !_inBlocks[i] ) fprintf( stderr , "[ERROR] Failed to allocate input block[%d]: %d\n" , i , XYBlocks.value+2 ) , exit( 0 );
	}
	_xBlocks = AllocPointer< Pointer( PixelI ) >( XYBlocks.value+2 );
	if( !_xBlocks ) fprintf( stderr , "[ERROR] Failed to allcoate x block: %d\n" , XYBlocks.value+2 ) , exit( 0 );
	_xyBlocks = AllocPointer< Pointer( PixelI ) >( XYBlocks.value );
	if( !_xyBlocks ) fprintf( stderr , "[ERROR] Failed to allcoate xy block: %d\n" , XYBlocks.value ) , exit( 0 );
	for( int j=0 ; j<XYBlocks.value ; j++ )
	{
		_xyBlocks[j] = NewPointer< PixelI >( bWidth * bHeight );
		if( !_xyBlocks[j] ) fprintf( stderr , "[ERROR] Failed to allocate xy block [%d]: %d x %d\n" , j , bWidth , bHeight ) , exit( 0 );
	}
	for( int j=0 ; j<XYBlocks.value+2 ; j++ )
	{
		_xBlocks[j] = NewPointer< PixelI >( bWidth * bHeight );
		if( !_xBlocks[j] ) fprintf( stderr , "[ERROR] Failed to allocate x block [%d] %d x %d\n" , j , bWidth , bHeight ) , exit( 0 );
		for( int i=0 ; i<3 ; i++ )
		{
			_inBlocks[i][j] = NewPointer< PixelI >( bWidth * bHeight );
			if( !_inBlocks[i][j] ) fprintf( stderr , "[ERROR] Failed to allocate input block [%d,%d]: %d x %d\n" , i , j , bWidth , bHeight ) , exit( 0 );
		}
	}

	ImageTileGrid3D* xySmoothed;
	{
		char** fileNames = new char*[ xBlocks * yBlocks * in->layers() ];
		for( int z=0 ; z<in->layers() ; z++ )
		{
			char fileHeader[512];
			if( MakeSubDirs.set )
			{
				sprintf( fileHeader , "%s_%d%c" , ScratchHeader.set ? ScratchHeader.value : "" , z , FILE_SEPARATOR );
				if( !MakeDir( fileHeader ) ) fprintf( stderr , "[ERROR] smoothXY: Failed to create directory: %s\n" , fileHeader );
			}
			else sprintf( fileHeader , "%s" , ScratchHeader.set ? ScratchHeader.value : "" );
			for( int _y=0 ; _y<yMultiBlocks ; _y++ )
			{
				// Read in the front blocks
#if PARALLEL_IO_CODE
#pragma omp parallel for num_threads( Threads.value )
#endif // PARALLEL_IO_CODE
				for( int j=-1 ; j<=XYBlocks.value ; j++ ) (*in)[z].readBlock( ( Pointer( unsigned char ) )_inBlocks[2][j+1] , 0 , _y*XYBlocks.value+j , Channels==1 );

				// Iterate over all the blocks in the row
				for( int x=0 ; x<xBlocks ; x++ )
				{
					if( Progress.set && omp_get_thread_num()==0 ) printf( "XY-Smoothing[%d %d] %d / %d               \r" , x , _y , z , in->layers() );
					// Circulate the blocks along the x-direction and read in the next block
					{
						for( int j=0 ; j<XYBlocks.value+2 ; j++ )
						{
							Pointer( PixelI ) temp = _inBlocks[0][j];
							_inBlocks[0][j] = _inBlocks[1][j];
							_inBlocks[1][j] = _inBlocks[2][j];
							_inBlocks[2][j] = temp;
						}
#if PARALLEL_IO_CODE
#pragma omp parallel for num_threads( Threads.value )
#endif // PARALLEL_IO_CODE
						for( int j=-1 ; j<=XYBlocks.value ; j++ ) (*in)[z].readBlock( ( Pointer( unsigned char ) )_inBlocks[2][j+1] , x+1 , _y*XYBlocks.value+j , Channels==1 );
					}
					int yOffset = _y*bHeight*XYBlocks.value , xOffset = x*bWidth;
					int jStart , jEnd;

					// Smooth across the x-direction (start with enough padding in the y-direction to support subsequent y-smoothing)
					jStart = std::max< int >( -yOffset , -xyRadius );
					jEnd = std::min< int >( in->height()-yOffset , bHeight*XYBlocks.value+xyRadius );
#pragma omp parallel for num_threads( Threads.value )
					for( int j=jStart ; j<jEnd ; j++ )
					{
						int yBlock = ( j + bHeight ) / bHeight , yIndex = ( j + bHeight ) - yBlock * bHeight;

						Pointer( PixelI ) xRow = _xBlocks[ yBlock ] + yIndex*bWidth;
						for( int i=0 ; i<bWidth && i<in->width()-xOffset ; i++ )
						{
							PixelF pSum;
							FLT sum = 0;
							for( int ii=i-xyRadius ; ii<=i+xyRadius ; ii++ )
								if( ii+xOffset<0 || ii+xOffset>in->width() ) continue;
								else
								{
									FLT weight = weights[ii-i+xyRadius];
									int xBlock = ( ii + bWidth ) / bWidth , xIndex = ii - ( xBlock - 1 ) * bWidth;
									PixelI p = _inBlocks[xBlock][yBlock][yIndex*bWidth+xIndex];
									sum += weight;
									pSum += ToFloat( p ) * weight;
								}
							xRow[i] = FromFloat( pSum / sum );
						}
					}
					// Smooth across the y-direction
					jStart = std::max< int >( -yOffset , 0 );
					jEnd = std::min< int >( in->height()-yOffset , bHeight*XYBlocks.value );
#pragma omp parallel for num_threads( Threads.value )
					for( int j=jStart ; j<jEnd ; j++ )
					{
						Pointer( PixelI ) xyRow;
						{
							int yBlock = j / bHeight , yIndex = j - yBlock * bHeight;
							xyRow = _xyBlocks[yBlock] + yIndex*bWidth;
						}
						for( int i=0 ; i<bWidth && i<in->width()-xOffset ; i++ )
						{
							PixelF pSum;
							FLT sum = 0;
							for( int jj=j-xyRadius ; jj<=j+xyRadius ; jj++ )
								if( jj+yOffset<0 || jj+yOffset>in->height() ) continue;
								else
								{
									FLT weight = weights[jj-j+xyRadius];
									int yBlock = ( jj + bHeight ) / bHeight , yIndex = ( jj + bHeight ) - yBlock * bHeight;
									PixelI p = _xBlocks[yBlock][yIndex*bWidth+i];
									sum += weight;
									pSum += ToFloat( p ) * weight;
								}
							xyRow[i] = FromFloat( pSum / sum );
						}
					}
					// Output the current tile
					{
						int _bWidth = std::min< int >( bWidth , in->width ()-x*bWidth  );
#if PARALLEL_IO_CODE
#pragma omp parallel for num_threads( Threads.value )
#else // !PARALLEL_IO_CODE
						char fileName[512];
#endif // PARALLEL_IO_CODE
						for( int yy=0 ; yy<XYBlocks.value ; yy++ )
						{
#if PARALLEL_IO_CODE
							char fileName[512];
#endif // PARALLEL_IO_CODE
							int yyy = _y*XYBlocks.value+yy;
							int _bHeight = std::min< int >( bHeight , in->height()-yyy*bHeight );
							if( yyy>=yBlocks ) continue;
							else
							{
								sprintf( fileName , "%s_xyscratch_z%d_y%d_x%d.%s" , fileHeader , z , yyy , x , OutExtension.set ? OutExtension.value : DEFAULT_EXTENSION );
								fileNames[ z*xBlocks*yBlocks + yyy*xBlocks + x ] = new char[ strlen( fileName ) + 1 ];
								strcpy( fileNames[ z*xBlocks*yBlocks + yyy*xBlocks + x ] , fileName );
								if( _bWidth<bWidth || _bHeight<bHeight )
								{
									Pointer( PixelI ) __xyBlock = AllocPointer< PixelI >( _bWidth * _bHeight );
									for( int j=0 ; j<_bHeight ; j++ ) for( int i=0 ; i<_bWidth ; i++ ) __xyBlock[j*_bWidth+i] = _xyBlocks[yy][j*bWidth+i];
									WriteImage< Channels >( fileName , __xyBlock , _bWidth , _bHeight );
									FreePointer( __xyBlock );
								}
								else WriteImage< Channels >( fileName , _xyBlocks[yy] , bWidth , bHeight );
							}
						}
					}
				}
			}
		}
		xySmoothed = new ImageTileGrid3D( fileNames , xBlocks , yBlocks , in->layers() );
		for( int i=0 ; i<xBlocks*yBlocks*in->layers() ; i++ ) delete[] fileNames[i];
		delete[] fileNames;
	}
	if( Progress.set ) printf( "\n" );
	for( int j=0 ; j<XYBlocks.value ; j++ ) DeletePointer( _xyBlocks[j] );
	DeletePointer( _xyBlocks );
	for( int j=0 ; j<XYBlocks.value+2 ; j++ )
	{
		DeletePointer( _xBlocks[j] );
		for( int i=0 ; i<3 ; i++ ) DeletePointer( _inBlocks[i][j] );
	}
	for( int i=0 ; i<3 ; i++ ) DeletePointer( _inBlocks[i] );
	DeletePointer( _xBlocks );
	delete[] weights;
	return xySmoothed;
}
#if LAZY_IO
template< int Channels >
ImageTileGrid3D* smoothZ( const ImageTileGrid3D* in )
{
	typedef Pixel< unsigned char , Channels > PixelI;
	typedef Pixel<           FLT , Channels > PixelF;
	PixelI black;

	int zRadius = (int)ceil( ZSigma.value*3 );
	int bWidth = in->bWidth() , bHeight = in->bHeight() , zSlices = 1 + 2*zRadius , _zSlices = zSlices + 1;

	FLT* weights = new FLT[zSlices];
	{
		FLT sum = 0;
		for( int i=-zRadius ; i<=zRadius ; i++ ) weights[i+zRadius] = (FLT)exp( -(double)(i*i)/( 2. * ZSigma.value * ZSigma.value ) ) , sum += weights[i+zRadius];
		for( int i=-zRadius ; i<=zRadius ; i++ ) weights[i+zRadius] /= sum;
	}
	int xBlocks = in->columns() , yBlocks = in->rows();
	if( Verbose.set ) printf( "\t\tBlock Size: %d x %d\n" , bWidth , bHeight ) , printf( "\t\tBlocks: %d x %d x %d\n" , xBlocks , yBlocks , in->layers() );
	Pointer( Pointer( PixelI ) ) inputBlocks = NewPointer< Pointer( PixelI ) >( _zSlices );
	if( !inputBlocks ) fprintf( stderr , "[ERRROR] smoothZ: Failed to allocate input blocks [%d]\n" , _zSlices ) , exit( 0 );
	Pointer( PixelI ) outputBlocks[2];
	for( int z=0 ; z<_zSlices ; z++ )
	{
		inputBlocks[z] = NewPointer< PixelI >( bWidth * bHeight );
		if( !inputBlocks[z] ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate input block (%d)[%d]\n" , z , bWidth * bHeight ) , exit( 0 );
	}
	if( MakeSubDirs.set )
		for( int z=0 ; z<in->layers() ; z++ )
		{
			char fileHeader[512];
			sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z , FILE_SEPARATOR );
			if( !MakeDir( fileHeader ) ) fprintf( stderr , "[ERROR] smoothZ: Failed to create directory: %s\n" , fileHeader );
		}

	for( int i=0 ; i<2 ; i++ )
	{
		outputBlocks[i] = NewPointer< PixelI >( bWidth * bHeight );
		if( !outputBlocks[i] ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate output block [%d]\n" , bWidth * bHeight ) , exit( 0 );
	}
	// Given P={p_1,\cdots,p_n}
	// var( P ) = [ \sum_i ( w_i * |p_i-q|^2 ) ] / \sum_i w_i		 w/ q = \sum_i w_i * p_i / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 + w_i * |q|^2 - 2 * w_i * < p_i , q > ) ] / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 ) +(\sum_i w_i) |q|^2 - 2 * < \sum_i w_i * p_i , q > ] / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 ) +(\sum_i w_i) |q|^2 ] / \sum_i w_i - 2 * |q|^2 ]
	Pointer( PixelF )  avgBlock = AllocPointer< PixelF >( bWidth*bHeight );
	Pointer( FLT )     sumBlock = AllocPointer< FLT >   ( bWidth*bHeight );
	Pointer( FLT )     varBlock = AllocPointer< FLT >   ( bWidth*bHeight );
	Pointer( PixelF ) _avgBlock = AllocPointer< PixelF >( bWidth*bHeight );
	if( !avgBlock || !sumBlock || !varBlock || !_avgBlock ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate memory for average calculation\n" ) , exit( 0 );

#define INPUT_INDEX( z ) ( ( (z)+_zSlices ) % _zSlices )
#define OUTPUT_INDEX( z )( ( (z)+2 ) % 2 )
	for( int y=0 ; y<yBlocks ; y++ ) for( int x=0 ; x<xBlocks ; x++ )
	{
		int yOffset = y*bHeight , xOffset = x*bWidth;
#if PARALLEL_IO_CODE
#pragma omp parallel for num_threads( Threads.value )
#endif // PARALLEL_IO_CODE
		for( int z=0 ; z<=zRadius ; z++ ) in->readBlock( ( Pointer( unsigned char ) )inputBlocks[ INPUT_INDEX(z) ] , x , y , z , Channels==1 );
		for( int z=0 ; z<=in->layers() ; z++ )
		{
#pragma omp parallel sections
			{
#pragma omp section
				{
					// Write out the old slice
					if( z-1>=0 )
					{
						Pointer( PixelI ) outputBlock = outputBlocks[ OUTPUT_INDEX(z-1) ];
						int _bWidth  = std::min< int >( bWidth  , in->width ()-x*bWidth  );
						int _bHeight = std::min< int >( bHeight , in->height()-y*bHeight );

						char fileHeader[512];
						if( MakeSubDirs.set ) sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z-1 , FILE_SEPARATOR );
						else                  sprintf( fileHeader , "%s" , OutHeader.set ? OutHeader.value : "" );

						char fileName[512];
						sprintf( fileName , "%s_z%d_y%d_x%d.%s" , fileHeader , z-1 , y , x , OutExtension.set ? OutExtension.value : DEFAULT_EXTENSION );
						if( _bWidth<bWidth || _bHeight<bHeight )
						{
							Pointer( PixelI ) __xyBlock = AllocPointer< PixelI >( _bWidth * _bHeight );
#pragma omp parallel for num_threads( Threads.value )
							for( int j=0 ; j<_bHeight ; j++ ) for( int i=0 ; i<_bWidth ; i++ ) __xyBlock[j*_bWidth+i] = outputBlock[ j*bWidth + i ];
							WriteImage< Channels >( fileName , __xyBlock , _bWidth , _bHeight );
							FreePointer( __xyBlock );
						}
						else WriteImage< Channels >( fileName , outputBlock , bWidth , bHeight );
					}
				}
#pragma omp section
				{
					// Read in the new slice
					if( z+1+zRadius<in->layers() ) in->readBlock( ( Pointer(unsigned char) )inputBlocks[ INPUT_INDEX( z+1+zRadius ) ] , x , y , z+1+zRadius , Channels==1 );
				}
				// Process the current slices
#pragma omp section
				{
					if( z<in->layers() )
					{
						if( Progress.set ) printf( "Z-Smoothing[%d %d] %d / %d                \r" , x , y , z , in->layers() );
						Pointer( PixelI ) inputBlock = inputBlocks[ INPUT_INDEX(z) ];
						Pointer( PixelI ) outputBlock = outputBlocks[ OUTPUT_INDEX(z) ];
						FLT sum = 0;
#pragma omp parallel for num_threads( Threads.value )
						for( int i=0 ; i<bWidth*bHeight ; i++ ) avgBlock[i] = PixelF() , varBlock[i] = sumBlock[i] = 0;

						// Estimate the weighted average and the variance of the pixel values
						for( int kk=-zRadius ; kk<=zRadius ; kk++ )
						{
							if( kk+z<0 || kk+z>=in->layers() ) continue;
							FLT weight = weights[kk+zRadius];
							ConstPointer( PixelI ) _inputBlock = inputBlocks[ INPUT_INDEX( z+kk ) ];
#pragma omp parallel for num_threads( Threads.value )
							for( int i=0 ; i<bWidth*bHeight ; i++ )
							{
								PixelF p = ToFloat( _inputBlock[i] );
								avgBlock[i] += p * weight;
								varBlock[i] += SquareNorm( p ) * weight;
								// Only average over the non-black pixels
								if( !( NoBlack.set && _inputBlock[i]==black ) ) sumBlock[i] += weight;
							}
						}

						// Iterate over the pixels once again, but this time only sum over pixels which are within ZRejectDeviation.value standard deviations of the current pixel's value
						// (i.e. pixels whose square distance from the weighted average is smaller than (ZRejectDeviation.value)^2x the variance)
						// If the center pixel is itself outside the standard deviation, keep its value fixed.
						if( !ZRejectDeviation.set || ZRejectDeviation.value<=0 )
#pragma omp parallel for num_threads( Threads.value )
							for( int i=0 ; i<bWidth*bHeight ; i++ ) _avgBlock[i] = avgBlock[i];
						else
						{
							// Divide through by the weight-sum to get the actual average/variance
#pragma omp parallel for num_threads( Threads.value )
							for( int i=0 ; i<bWidth*bHeight ; i++ )
							{
								if( sumBlock[i]!=0 )
								{
									avgBlock[i] /= sumBlock[i];
									varBlock[i] /= sumBlock[i];
								}
								else
								{
									avgBlock[i] = ToFloat( black );
									varBlock[i] = 0;
								}
								varBlock[i] -= SquareNorm( avgBlock[i] );
								varBlock[i] *= ZRejectDeviation.value * ZRejectDeviation.value;
								sumBlock[i] = 0;
								_avgBlock[i] = PixelF();
							}
							// If the center value is too far form the average, then keep the center's value fixed
							{
#pragma omp parallel for num_threads( Threads.value )
								for( int i=0 ; i<bWidth*bHeight ; i++ )
								{
									// Check if the pixel is within the appropriate number of standard deviations
									PixelF p = ToFloat( inputBlock[i] );
									if( SquareNorm( p-avgBlock[i] )>varBlock[i] ) avgBlock[i] = p , varBlock[i] = 0;
								}
							}
							// Iterate over all contributing pixels and if the variance is low enough, incoporate them in the average
							for( int kk=-zRadius ; kk<=zRadius ; kk++ )
							{
								if( kk+z<0 || kk+z>=in->layers() ) continue;
								FLT weight = weights[kk+zRadius];
								ConstPointer( PixelI ) _inputBlock = inputBlocks[ INPUT_INDEX( z + kk ) ];
#pragma omp parallel for num_threads( Threads.value )
								for( int i=0 ; i<bWidth*bHeight ; i++ )
								{
									PixelF p = ToFloat( _inputBlock[i] );
									if( SquareNorm( p-avgBlock[i] )<=varBlock[i] ) sumBlock[i] += weight , _avgBlock[i] += p * weight;
								}
							}
						}

#pragma omp parallel for num_threads( Threads.value )
						for( int i=0 ; i<bWidth*bHeight ; i++ )
							if( ( NoBlack.set && inputBlock[i]==black ) || sumBlock[i]==0 ) outputBlock[i] = black;
							else                                                            outputBlock[i] = FromFloat( _avgBlock[i] / sumBlock[i] );
					}
				}
			}
		}
	}
	if( Progress.set ) printf( "\n" );
	for( int i=0 ; i<_zSlices ; i++ ) DeletePointer( inputBlocks[i] );
	for( int i=0 ; i<2 ; i++ ) DeletePointer( outputBlocks[i] );
	FreePointer(  avgBlock );
	FreePointer(  sumBlock );
	FreePointer(  varBlock );
	FreePointer( _avgBlock );

	delete[] weights;
	if( Out.set )
	{
		FILE* fp = fopen( Out.value , "w" );
		if( !fp ) fprintf( stderr , "[ERROR] Failed to open file for writing: %s\n" , Out.value ) , exit( 0 );
		fprintf( fp , "width: %d\n" , xBlocks );
		fprintf( fp , "height: %d\n" , yBlocks );
		fprintf( fp , "depth: %d\n" , in->layers() );
		for( int z=0 ; z<in->layers() ; z++ )
		{
			char fileHeader[512];
			if( MakeSubDirs.set ) sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z , FILE_SEPARATOR );
			else                  sprintf( fileHeader , "%s" , OutHeader.set ? OutHeader.value : "" );
			for( int y=0 ; y<yBlocks ; y++ ) for( int x=0 ; x<xBlocks ; x++ ) fprintf( fp , "%s_z%d_y%d_x%d.%s\n" , fileHeader , z , y , x , OutExtension.set ? OutExtension.value : DEFAULT_EXTENSION );
		}
		fclose( fp );
	}
}
#else // !LAZY_IO
template< int Channels >
ImageTileGrid3D* smoothZ( const ImageTileGrid3D* in )
{
	typedef Pixel< unsigned char , Channels > PixelI;
	typedef Pixel<           FLT , Channels > PixelF;
	PixelI black;

	int zRadius = (int)ceil( ZSigma.value*3 );
	int bWidth = in->bWidth() , bHeight = in->bHeight() , zSlices = 1 + 2*zRadius;

	FLT* weights = new FLT[zSlices];
	{
		FLT sum = 0;
		for( int i=-zRadius ; i<=zRadius ; i++ ) weights[i+zRadius] = (FLT)exp( -(double)(i*i)/( 2. * ZSigma.value * ZSigma.value ) ) , sum += weights[i+zRadius];
		for( int i=-zRadius ; i<=zRadius ; i++ ) weights[i+zRadius] /= sum;
	}
	int xBlocks = in->columns() , yBlocks = in->rows();
	if( Verbose.set ) printf( "\t\tBlock Size: %d x %d\n" , bWidth , bHeight ) , printf( "\t\tBlocks: %d x %d x %d\n" , xBlocks , yBlocks , in->layers() );
	Pointer( Pointer( PixelI ) ) volumeBlocks = NewPointer< Pointer( PixelI ) >( zSlices );
	if( !volumeBlocks ) fprintf( stderr , "[ERRROR] smoothZ: Failed to allocate volume blocks [%d]\n" , zSlices ) , exit( 0 );
	Pointer( PixelI ) outputBlock;
	for( int z=0 ; z<zSlices ; z++ )
	{
		volumeBlocks[z] = NewPointer< PixelI >( bWidth * bHeight );
		if( !volumeBlocks[z] ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate volume block (%d)[%d]\n" , z , bWidth * bHeight ) , exit( 0 );
	}
	if( MakeSubDirs.set )
		for( int z=0 ; z<in->layers() ; z++ )
		{
			char fileHeader[512];
			sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z , FILE_SEPARATOR );
			if( !MakeDir( fileHeader ) ) fprintf( stderr , "[ERROR] smoothZ: Failed to create directory: %s\n" , fileHeader );
		}

	outputBlock = NewPointer< PixelI >( bWidth * bHeight );
	if( !outputBlock ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate output block [%d]\n" , bWidth * bHeight ) , exit( 0 );
	// Given P = { p_1 , \cdots , p_n }
	// var( P ) = [ \sum_i ( w_i * |p_i-q|^2 ) ] / \sum_i w_i		 w/ q = \sum_i w_i * p_i / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 + w_i * |q|^2 - 2 * w_i * < p_i , q > ) ] / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 ) +(\sum_i w_i) |q|^2 - 2 * < \sum_i w_i * p_i , q > ] / \sum_i w_i
	//          = [ \sum_i ( w_i * |p_i|^2 ) +(\sum_i w_i) |q|^2 ] / \sum_i w_i - 2 * |q|^2 ]
	Pointer( PixelF )  avgBlock = AllocPointer< PixelF >( bWidth*bHeight );
	Pointer( FLT )     sumBlock = AllocPointer< FLT >   ( bWidth*bHeight );
	Pointer( FLT )     varBlock = AllocPointer< FLT >   ( bWidth*bHeight );
	Pointer( PixelF ) _avgBlock = AllocPointer< PixelF >( bWidth*bHeight );
	if( !avgBlock || !sumBlock || !varBlock || !_avgBlock ) fprintf( stderr , "[ERROR] smoothZ: Failed to allocate memory for average calculation\n" ) , exit( 0 );

	for( int y=0 ; y<yBlocks ; y++ ) for( int x=0 ; x<xBlocks ; x++ )
	{
		int yOffset = y*bHeight , xOffset = x*bWidth;
#if PARALLEL_IO_CODE
#pragma omp parallel for num_threads( Threads.value )
#endif // PARALLEL_IO_CODE
		for( int z=0 ; z<zRadius ; z++ ) in->readBlock( ( Pointer( unsigned char ) )volumeBlocks[z+1] , x , y , z , Channels==1 );
		for( int z=0 ; z<in->layers() ; z++ )
		{
			// Circulate the blocks along the z-direction and read in the next block
			{
				Pointer( PixelI ) temp = volumeBlocks[0];
				for( int zz=0 ; zz<zSlices-1 ; zz++ ) volumeBlocks[zz] = volumeBlocks[zz+1];
				volumeBlocks[zSlices-1] = temp;
				in->readBlock( ( Pointer(unsigned char) )volumeBlocks[zRadius] , x , y , z+zRadius , Channels==1 );
			}

			if( Progress.set ) printf( "Z-Smoothing[%d %d] %d / %d                \r" , x , y , z , in->layers() );
			FLT sum = 0;
#pragma omp parallel for num_threads( Threads.value )
			for( int i=0 ; i<bWidth*bHeight ; i++ ) avgBlock[i] = PixelF() , varBlock[i] = sumBlock[i] = 0;

			// Estimate the weighted average and the variance of the pixel values
			for( int kk=-zRadius ; kk<=zRadius ; kk++ )
			{
				if( kk+z<0 || kk+z>=in->layers() ) continue;
				FLT weight = weights[kk+zRadius];
				ConstPointer( PixelI ) volumeBlock = volumeBlocks[ ( kk + zSlices ) % zSlices ];
#pragma omp parallel for num_threads( Threads.value )
				for( int i=0 ; i<bWidth*bHeight ; i++ )
				{
					PixelF p = ToFloat( volumeBlock[i] );
					avgBlock[i] += p * weight;
					varBlock[i] += SquareNorm( p ) * weight;
					// Only average over the non-black pixels
					if( !( NoBlack.set && volumeBlock[i]==black ) ) sumBlock[i] += weight;
				}
			}

			// Iterate over the pixels once again, but this time only sum over pixels which are within ZRejectDeviation.value standard deviations of the current pixel's value
			// (i.e. pixels whose square distance from the weighted average is smaller than (ZRejectDeviation.value)^2x the variance)
			// If the center pixel is itself outside the standard deviation, keep its value fixed.
			if( !ZRejectDeviation.set || ZRejectDeviation.value<=0 )
#pragma omp parallel for num_threads( Threads.value )
				for( int i=0 ; i<bWidth*bHeight ; i++ ) _avgBlock[i] = avgBlock[i];
			else
			{
				// Divide through by the weight-sum to get the actual average/variance
#pragma omp parallel for num_threads( Threads.value )
				for( int i=0 ; i<bWidth*bHeight ; i++ )
				{
					if( sumBlock[i]!=0 )
					{
						avgBlock[i] /= sumBlock[i];
						varBlock[i] /= sumBlock[i];
					}
					else
					{
						avgBlock[i] = ToFloat( black );
						varBlock[i] = 0;
					}
					varBlock[i] -= SquareNorm( avgBlock[i] );
					varBlock[i] *= ZRejectDeviation.value * ZRejectDeviation.value;
					sumBlock[i] = 0;
					_avgBlock[i] = PixelF();
				}
				// If the center value is too far form the average, then keep the center's value fixed
				{
#pragma omp parallel for num_threads( Threads.value )
					for( int i=0 ; i<bWidth*bHeight ; i++ )
					{
						// Check if the pixel is within the appropriate number of standard deviations
						PixelF p = ToFloat( volumeBlocks[0][i] );
						if( SquareNorm( p-avgBlock[i] )>varBlock[i] ) avgBlock[i] = p , varBlock[i] = 0;
					}
				}
				// Iterate over all contributing pixels and if the variance is low enough, incoporate them in the average
				for( int kk=-zRadius ; kk<=zRadius ; kk++ )
				{
					if( kk+z<0 || kk+z>=in->layers() ) continue;
					FLT weight = weights[kk+zRadius];
					ConstPointer( PixelI ) volumeBlock = volumeBlocks[ ( kk + zSlices ) % zSlices ];
#pragma omp parallel for num_threads( Threads.value )
					for( int i=0 ; i<bWidth*bHeight ; i++ )
					{
						PixelF p = ToFloat( volumeBlock[i] );
						if( SquareNorm( p-avgBlock[i] )<=varBlock[i] ) sumBlock[i] += weight , _avgBlock[i] += p * weight;
					}
				}
			}

#pragma omp parallel for num_threads( Threads.value )
			for( int i=0 ; i<bWidth*bHeight ; i++ )
				if( ( NoBlack.set && volumeBlocks[0][i]==black ) || sumBlock[i]==0 ) outputBlock[i] = black;
				else                                                                 outputBlock[i] = FromFloat( _avgBlock[i] / sumBlock[i] );

			// Write out the smoothed tiles
			{
				int _bWidth  = std::min< int >( bWidth  , in->width ()-x*bWidth  );
				int _bHeight = std::min< int >( bHeight , in->height()-y*bHeight );

				char fileHeader[512];
				if( MakeSubDirs.set ) sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z , FILE_SEPARATOR );
				else                  sprintf( fileHeader , "%s" , OutHeader.set ? OutHeader.value : "" );

				char fileName[512];
				sprintf( fileName , "%s_z%d_y%d_x%d.%s" , fileHeader , z , y , x , OutExtension.set ? OutExtension.value : DEFAULT_EXTENSION );
				if( _bWidth<bWidth || _bHeight<bHeight )
				{
					Pointer( PixelI ) __xyBlock = AllocPointer< PixelI >( _bWidth * _bHeight );
#pragma omp parallel for num_threads( Threads.value )
					for( int j=0 ; j<_bHeight ; j++ ) for( int i=0 ; i<_bWidth ; i++ ) __xyBlock[j*_bWidth+i] = outputBlock[ j*bWidth + i ];
					WriteImage< Channels >( fileName , __xyBlock , _bWidth , _bHeight );
					FreePointer( __xyBlock );
				}
				else WriteImage< Channels >( fileName , outputBlock , bWidth , bHeight );
			}
		}
	}
	if( Progress.set ) printf( "\n" );
	for( int i=0 ; i<zSlices ; i++ ) DeletePointer( volumeBlocks[i] );
	DeletePointer( outputBlock );
	FreePointer(  avgBlock );
	FreePointer(  sumBlock );
	FreePointer(  varBlock );
	FreePointer( _avgBlock );

	delete[] weights;
	if( Out.set )
	{
		FILE* fp = fopen( Out.value , "w" );
		if( !fp ) fprintf( stderr , "[ERROR] Failed to open file for writing: %s\n" , Out.value ) , exit( 0 );
		fprintf( fp , "width: %d\n" , xBlocks );
		fprintf( fp , "height: %d\n" , yBlocks );
		fprintf( fp , "depth: %d\n" , in->layers() );
		for( int z=0 ; z<in->layers() ; z++ )
		{
			char fileHeader[512];
			if( MakeSubDirs.set ) sprintf( fileHeader , "%s_%d%c" , OutHeader.set ? OutHeader.value : "" , z , FILE_SEPARATOR );
			else                  sprintf( fileHeader , "%s" , OutHeader.set ? OutHeader.value : "" );
			for( int y=0 ; y<yBlocks ; y++ ) for( int x=0 ; x<xBlocks ; x++ ) fprintf( fp , "%s_z%d_y%d_x%d.%s\n" , fileHeader , z , y , x , OutExtension.set ? OutExtension.value : DEFAULT_EXTENSION );
		}
		fclose( fp );
	}
}
#endif // LAZY_IO
int main( int argc , char* argv[] )
{
	cmdLineParse( argc-1 , argv+1 , sizeof( params ) / sizeof( cmdLineReadable*) , params );
	if( !In.set )
	{
		ShowUsage( argv[0] );
		return EXIT_FAILURE;
	}
	XYBlocks.value = std::max< int >( XYBlocks.value , 1 );
	if( Channels.set && Channels.value!=1 && Channels.value!=3 ) fprintf( stderr , "[WARNING] Only 1 and 3 channels supported. Defaulting to input type.\n" ) , Channels.set = false;

	ImageTileGrid3D ti( In.value );
	if( Verbose.set )
	{
		size_t current , peak;
		WorkingSetInfo( current , peak );
		printf( "Current/Peak Memory (MB): %d %d\n" , current>>20 , peak>>20 );
	}

	bool gray;
	if( Channels.set ) gray = ( Channels.value==1 );
	else               gray = ti.gray();

	double t = Time() , _t = t;
	ImageTileGrid3D* xySmoothed;
	if( gray ) xySmoothed = smoothXY< 1 >( &ti );
	else       xySmoothed = smoothXY< 3 >( &ti );
	if( Verbose.set )
	{
		printf( "\tXY Time: %.2f(s)\n" , Time()-_t ) , _t = Time();
		size_t current , peak;
		WorkingSetInfo( current , peak );
		printf( "\tPeak Memory (MB): %d\n" , peak>>20 );
	}
	ImageTileGrid3D* xyzSmoothed;
	if( gray ) xyzSmoothed = smoothZ< 1 >( xySmoothed );
	else       xyzSmoothed = smoothZ< 3 >( xySmoothed );
	if( Verbose.set )
	{
		printf( "\t Z Time: %.2f(s)\n" , Time()-_t ) , _t = Time();
		size_t current , peak;
		WorkingSetInfo( current , peak );
		printf( "Peak Memory (MB): %d\n" , peak>>20 );
	}
	if( !NoCleanUp.set ) xySmoothed->removeFiles();
	delete xySmoothed;
	if( Verbose.set )
	{
		printf( "Time: %.2f(s)\n" , Time()-t );
		size_t current , peak;
		WorkingSetInfo( current , peak );
		printf( "Peak Memory (MB): %d\n" , peak>>20 );
	}
	return EXIT_SUCCESS;
}
